from typing import List, NamedTuple


class VocabElement(NamedTuple):
    token: str
    frequency: int
    proportion: float
    range: int
    stdev: float
    vc: float
    juilland_d: float
    carroll_d2: float
    rosengren_sadj: float
    dp: float
    dp_norm: float
    kl_divergence: float


CorpusVocabulary = List[VocabElement]



import math
from numbers import Number
from typing import Dict, Union

import numpy as np
from scipy.sparse import csc_matrix
from scipy.special import rel_entr
from scipy.stats import entropy
from tqdm import tqdm

from .corpustypes import CorpusVocabulary, VocabElement

NATSPERBIT = np.log2(math.e)


def natstobits(nats: float) -> float:
    return nats * NATSPERBIT


def retreive_scalar(value: Union[np.ndarray, float, int]) -> Union[float, int]:
    """Helper function to get scalars out of single-element arrays.
    Also runds to help with floating point"""
    if isinstance(value, np.ndarray) and value.shape == (1,):
        return np.round(value[0], 15)
    elif isinstance(value, Number):
        return np.round(value, 15)
    else:
        raise ValueError(
            f"{value} not Number or single element array - it's a {type(value)}"
        )


def calculate_all_stats(
    tdm: csc_matrix, vocabulary: Dict[str, int]
) -> CorpusVocabulary:
    doc_lengths = tdm.sum(axis=1)
    corpus_len = tdm.shape[0]
    token_counts = tdm.sum()
    s = doc_lengths / token_counts
    range_array = tdm.getnnz(axis=0)

    stats = []
    pbar_columns = tqdm(
        range(tdm.shape[1]), desc="Calculating Staistics for Each Token"
    )
    for column, token in zip(pbar_columns, vocabulary):
        # Imprecision in some calculations without float64 type
        col_d = tdm.getcol(column).toarray()
        freq = col_d.sum()
        prop = freq / token_counts
        range_ = range_array[column]
        stdev = np.std(col_d)
        vc = stdev / col_d.mean()
        # p = proportion of doc comprised of this token
        # s = percent of corpus comprised of tokens in this doc
        # v = percent of tokens in doc that are this token
        v = col_d / freq
        p = col_d / doc_lengths
        p_std = np.std(p)
        p_mean = np.mean(p)
        juilland_d = 1 - (p_std / p_mean) * (1 / np.sqrt(corpus_len - 1))
        carroll_d2 = entropy(p, base=2) / np.log2(corpus_len)
        rosengren_sadj = ((np.sqrt(np.multiply(s, col_d))).sum() ** 2) / freq
        dp = (np.abs(np.subtract(v, s))).sum() * 0.5
        dp_norm = dp / (1 - np.min(s))
        kl = natstobits(rel_entr(v, s).sum())
        element = VocabElement(
            token=token,
            frequency=retreive_scalar(freq),
            proportion=retreive_scalar(prop),
            range=retreive_scalar(range_),
            stdev=retreive_scalar(stdev),
            vc=retreive_scalar(vc),
            juilland_d=retreive_scalar(juilland_d),
            carroll_d2=retreive_scalar(carroll_d2),
            rosengren_sadj=retreive_scalar(rosengren_sadj),
            dp=retreive_scalar(dp),
            dp_norm=retreive_scalar(dp_norm),
            kl_divergence=retreive_scalar(kl),
        )
        stats.append(element)
    return stats



from collections import Counter
from functools import cached_property
from pathlib import Path
from typing import FrozenSet, List, Optional

import srsly
from spacy.language import Language
from spacy.tokens import Doc


@Language.factory("simple_corpus_stats", default_config={"n_train": None})
def create_simple_corpus_stats_component(
    nlp: Language, name: str, n_train: Optional[int]
):
    return SimpleCorpusStatistics(nlp, n_train=n_train)


class SimpleCorpusStatistics:
    def __init__(self, nlp: Language, n_train: Optional[int]):
        self.vocabulary: Counter = Counter()
        self.doc_lengths: List[int] = []
        self.n_train = n_train

        self._known_length = n_train is not None
        self._call_count: int = 0

    def __call__(self, doc: Doc) -> Doc:
        if self._known_length and self._call_count >= self.n_train:
            return doc

        tokens = [token.text for token in doc]
        self.vocabulary.update(tokens)
        self.doc_lengths.append(len(tokens))
        self._call_count += 1
        return doc

    def __getitem__(self, key: str) -> int:
        return self.vocabulary[key]

    def __contains__(self, key: str) -> int:
        return key in self.vocabulary

    def __len__(self) -> int:
        return len(self.vocabulary)

    @cached_property
    def vocab_size(self) -> int:
        return len(self)

    @cached_property
    def token_count(self) -> int:
        return sum(self.doc_lengths)

    @cached_property
    def corpus_length(self) -> int:
        return len(self.doc_lengths)

    @cached_property
    def type_token_ratio(self) -> float:
        return self.vocab_size / self.token_count

    @cached_property
    def hapax_legomena(self) -> FrozenSet:
        return frozenset(
            [token for token, count in self.vocabulary.items() if count == 1]
        )

    @cached_property
    def dis_legomena(self) -> FrozenSet:
        return frozenset(
            [token for token, count in self.vocabulary.items() if count == 2]
        )

    def to_disk(self, path, exclude=tuple()):
        if not isinstance(path, Path):
            path = Path(path)
        if not path.exists():
            path.mkdir()
        srsly.write_msgpack(path / "vocabulary.msgpack", dict(self.vocabulary))
        srsly.write_msgpack(path / "doc_lengths.msgpack", self.doc_lengths)
        srsly.write_json(
            path / "vocabulary-meta.json",
            dict(
                n_train=self.n_train,
                _call_count=self._call_count,
            ),
        )

    def from_disk(self, path, exclude=tuple()):
        self.vocabulary = Counter(srsly.read_msgpack(path / "vocabulary.msgpack"))
        self.doc_lengths = srsly.read_msgpack(path / "doc_lengths.msgpack")
        meta = srsly.read_json(path / "vocabulary-meta.json")
        self.n_train = meta["n_train"]
        self._call_count = meta["_call_count"]
        return self



import array
import functools
import operator
from collections import Counter, defaultdict
from functools import cached_property
from pathlib import Path
from typing import Callable, Dict, List, NamedTuple, Union

import numpy as np
import scipy.sparse as sp
import srsly
from spacy.language import Language
from spacy.tokens import Doc

from .corpustypes import CorpusVocabulary, VocabElement
from .dispersion_functions import calculate_all_stats


def require_frozen(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.frozen:
            raise ValueError("Must call freeze() before use")
        return func(self, *args, **kwargs)

    return wrapper


@Language.factory("corpus_statistics", default_config={"lowercase": False})
def create_corpus_statistics_component(nlp: Language, name: str, lowercase: bool):
    return CorpusStatistics(nlp, lowercase=lowercase)


class CorpusAggregates(NamedTuple):
    vocab_size: int
    token_count: int
    corpus_length: int
    doc_lengths: List[int]


class CorpusStatistics:
    def __init__(self, nlp: Language, lowercase: bool):
        self.corpus_counter: Counter = Counter()
        self.lowercase = lowercase

        self.vocabulary: Dict[str, int] = defaultdict()
        self.vocabulary.default_factory = self.vocabulary.__len__

        self._j_indices: List[int] = []
        self._indptr: List[int] = [0]
        self._values = array.array("i")

        self.frozen: bool = False

    def __call__(self, doc: Doc) -> Doc:
        # TODO: Raise Error if Frozen
        feature_counter: Dict[int, int] = {}
        for token in doc:
            token_text = token.text if not self.lowercase else token.text.lower()
            feature_idx = self.vocabulary[token_text]
            if feature_idx not in feature_counter:
                feature_counter[feature_idx] = 1
            else:
                feature_counter[feature_idx] += 1
        self._j_indices.extend(feature_counter.keys())
        self._values.extend(feature_counter.values())
        self._indptr.append(len(self._j_indices))
        return doc

    def freeze(self) -> None:
        j_indices = np.asarray(self._j_indices, dtype=np.int64)
        indptr = np.asarray(self._indptr, dtype=np.int64)
        values = np.frombuffer(self._values, dtype=np.intc)
        _X = sp.csr_matrix(
            (values, j_indices, indptr),
            shape=(len(indptr) - 1, len(self.vocabulary)),
            dtype=np.int64,
        )
        _X.sort_indices()
        _X = _X.tocsc()

        # freeze vocabulary: convert from defaultdict
        self.vocabulary = dict(self.vocabulary)
        self.aggregates = CorpusAggregates(
            vocab_size=len(self.vocabulary),
            token_count=_X.sum(),
            corpus_length=_X.shape[0],
            doc_lengths=_X.sum(axis=1).ravel().tolist(),
        )

        self.data = calculate_all_stats(_X, self.vocabulary)
        self.frozen = True

    @require_frozen
    def __getitem__(self, key: str) -> VocabElement:
        element = [v for v in self.data if v.token == key]
        if element:
            return element[0]
        else:
            raise KeyError(f"{key} not in vocabulary.")

    @require_frozen
    def __contains__(self, key: str) -> int:
        return key in self.vocabulary

    def __len__(self):
        raise ValueError(
            "Length is ambiguous. "
            "Do you mean vocab_size, token_count, or corpus_length?"
        )

    @cached_property  # type: ignore
    @require_frozen
    def vocab_size(self) -> int:
        return self.aggregates.vocab_size

    @cached_property  # type: ignore
    @require_frozen
    def token_count(self) -> int:
        return self.aggregates.token_count

    @cached_property  # type: ignore
    @require_frozen
    def corpus_length(self) -> int:
        return self.aggregates.corpus_length

    @cached_property  # type: ignore
    @require_frozen
    def doc_lengths(self) -> List[int]:
        return self.aggregates.doc_lengths

    @cached_property  # type: ignore
    @require_frozen
    def type_token_ratio(self) -> float:
        return self.aggregates.vocab_size / self.aggregates.token_count

    @require_frozen
    def query(
        self,
        attribute: str,
        op: Callable[[float, float], bool],
        value: Union[float, int],
    ) -> CorpusVocabulary:
        return [v for v in self.data if op(getattr(v, attribute), value)]

    @cached_property  # type: ignore
    @require_frozen
    def hapax_legomena(self) -> CorpusVocabulary:
        return self.query("frequency", operator.eq, 1)

    @cached_property  # type: ignore
    @require_frozen
    def dis_legomena(self) -> CorpusVocabulary:
        return self.query("frequency", operator.eq, 2)

    @require_frozen
    def to_disk(self, path, exclude=tuple()):
        if not isinstance(path, Path):
            path = Path(path)
        if not path.exists():
            path.mkdir()
        srsly.write_msgpack(path / "vocabulary.msgpack", self.vocabulary)
        srsly.write_msgpack(path / "dispersion_data.msgpack", self.data)
        srsly.write_msgpack(path / "aggregates.msgpack", self.aggregates)

    def from_disk(self, path, exclude=tuple()):
        self.vocabulary = srsly.read_msgpack(path / "vocabulary.msgpack")
        self.data = srsly.read_msgpack(path / "dispersion_data.msgpack")
        self.aggregates = srsly.read_msgpack(path / "aggregates.msgpack")
        self.frozen = True
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['corpus_statistics']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22.2,<2.0.0',
 'scipy>=1.8.0,<2.0.0',
 'spacy>=3.2.2,<4.0.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'spacy_factories': ['simple_corpus_stats = '
                     'corpus_statistics.component_simple:create_simple_corpus_stats_component']}

setup_kwargs = {
    'name': 'corpus-statistics',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Corpus Statistics: A Very Basic (for now) spaCy Pipeline Component\n\nIf you want to know what tokens your pipeline has seen, this is the component for you.\n\n```bash\npip install git+https://github.com/pmbaumgartner/corpus_statistics\n```\n\n⚡️ **Example**\n\n```python\nfrom spacy.lang.en import English\n\n# Use some example data\nfrom datasets import load_dataset\ndataset = load_dataset("imdb")\ntexts = dataset["train"]["text"]\n\n# ✨ start the magic \nnlp = English()  # or spacy.load(\'a_model\')\nnlp.add_pipe("simple_corpus_stats")\n\nfor doc in nlp.pipe(texts):\n    # ➡️ do your pipeline stuff! ➡️\n    pass\n\ncorpus_stats = nlp.get_pipe("simple_corpus_stats")\n\n# check if a token has been processed through this pipeline\ntoken = "apple"\nif token in corpus_stats:\n    token_count = corpus_stats[token]\n    print(f"\'{token}\' mentioned {token_count} times")\n\n# \'apple\' mentioned 24 times\n```\n\nIt\'s got all your favorite [legomena](https://en.wikipedia.org/wiki/Hapax_legomenon) like `hapax` and `dis`.\n\n```python\nonly_seen_once = len(corpus_stats.hapax_legomena)\npercent_of_vocab = only_seen_once / corpus_stats.vocab_size\nprint(f"{percent_of_vocab*100:.1f}% tokens only occurred once.")\n# 47.6% tokens only occurred once.\n\nonly_seen_twice = len(corpus_stats.dis_legomena)\npercent_of_vocab_2x = only_seen_twice / corpus_stats.vocab_size\nprint(f"{percent_of_vocab_2x*100:.1f}% tokens occurred twice.")\n# 12.3% tokens occurred twice.\n```\n\nWe counted some things too:\n\n```python\n# corpus_stats.vocabulary is a collections.Counter 🔢\nprint(*corpus_stats.vocabulary.most_common(5), sep="\\n")\n# (\'the\', 289838)\n# (\',\', 275296)\n# (\'.\', 236702)\n# (\'and\', 156484)\n# (\'a\', 156282)\n\nmean_doc_length = sum(corpus_stats.doc_lengths) / corpus_stats.corpus_length\nprint(f"Mean doc length: {mean_doc_length:.1f}")\n# Mean doc length: 272.5\n```\n\n# Use in Model Training and Config Files\n\nThis can be quite helpful if you wanted to know what tokens were seen in your training data. You can include this component in your training config as follows.\n\n```\n...\n[nlp]\nlang = "en"\npipeline = ["simple_corpus_stats", ...]\n...\n\n[components]\n\n[components.simple_corpus_stats]\nfactory = "simple_corpus_stats"\nn_train = 1000  # This is important! See below\n```\n\n⚠️ 🔁 If you use this component in a training config, your pipeline will see the same docs multiple times, due to the number of training epochs and evaluation steps, so the vocab counter will be incorrect. To correct for this, you need to specify the number of examples in your training dataset as the `n_train` config parameter. \n\n```python\nimport spacy\n\nnlp = spacy.load("your_trained_model")\ncorpus_stats = nlp.get_pipe("simple_corpus_stats")\n\nassert min(corpus_stats.vocabulary.values()) == 1\n\n# value from config\nassert len(corpus_stats.doc_lengths) == 1000\n```\n\n\n',
    'author': 'Peter Baumgartner',
    'author_email': '5107405+pmbaumgartner@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
